/*
 * ppmdistort version 0.3
 *  correct distortion of photo
 *  1996/07/15 version 0.1 by Masahide Kashiwagi
 *  1996/08/15 -DSAVE_MEMORY for saving memory
 *             -DDOS for compiling on MS-DOS
 *  1996/08/16 windows BMP file
 *             version 0.2
 *  1996/08/25 compile on LSI-C 86
 *             lcc -O -DDOS -DLSI_C ppmdstrt.c -ltinymain.obj -lintlib
 *             version 0.3
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#ifdef DOS
#ifdef LSI_C
#include <dos.h> /* farmalloc */
#else
#include <fcntl.h> /* O_BINARY */
#include <io.h> /* setmode */
#endif
#endif /* DOS */

#ifdef DOS
#ifndef SAVE_MEMORY
#define SAVE_MEMORY
#endif
#endif /* DOS */

#define FAR_PTR
#if defined(DOS) && defined(LSI_C)
#undef FAR_PTR
#define FAR_PTR far
#define malloc farmalloc
#define free farfree
#endif


typedef struct {
	unsigned char r;
	unsigned char g;
	unsigned char b;
} rgb;

int xsize, ysize;
unsigned char FAR_PTR * FAR_PTR *dat;

char *progname;

int full = 0;
int no_anti = 0;
double param = 20;


rgb get_rgb2(int x, int y)
{
	rgb color;
	unsigned char FAR_PTR *p;

	if (x < 0 || x >= xsize || y < 0 || y >= ysize) {
		color.r = color.g = color.b = 0;
	} else {
		p = dat[y] + x*3;
		color.r = *(p++);
		color.g = *(p++);
		color.b = *p;
	}

	return color;
}

rgb get_rgb(double x, double y)
{
	double r, g, b;
	double x_wait, y_wait;
	int xx, yy;
	rgb color;

	if (no_anti == 1) {
		return get_rgb2((int)(x+.5), (int)(y+.5));
	}

	r = g = b = 0.;
	xx = (int)x;
	yy = (int)y;

	x_wait = x - (double)xx;
	y_wait = y - (double)yy;

	color = get_rgb2(xx,yy);
	r += color.r * (1-x_wait) * (1-y_wait);
	g += color.g * (1-x_wait) * (1-y_wait);
	b += color.b * (1-x_wait) * (1-y_wait);
	color = get_rgb2(xx+1,yy);
	r += color.r * (x_wait) * (1-y_wait);
	g += color.g * (x_wait) * (1-y_wait);
	b += color.b * (x_wait) * (1-y_wait);
	color = get_rgb2(xx,yy+1);
	r += color.r * (1-x_wait) * (y_wait);
	g += color.g * (1-x_wait) * (y_wait);
	b += color.b * (1-x_wait) * (y_wait);
	color = get_rgb2(xx+1,yy+1);
	r += color.r * (x_wait) * (y_wait);
	g += color.g * (x_wait) * (y_wait);
	b += color.b * (x_wait) * (y_wait);
	
	color.r = (unsigned char)(r + .5);
	color.g = (unsigned char)(g + .5);
	color.b = (unsigned char)(b + .5);

	return color;
}


void skip_blank(FILE *fp)
{
	int c;
	while (1) {
		c = getc(fp);
		if (isspace(c)) {
			putchar(c);
			continue;
		}
		if (c != '#') break;
		putchar(c);

		while (1) {
			c = putchar(getc(fp));
			if (c == '\n') break;
		}
	}
	ungetc(c, fp);
}

int get_num(FILE *fp)
{
	int r, c;

	r = 0;
	while (1) {
		c = getc(fp);
		if (!isdigit(c)) break;
		r = r*10 + c - '0';
		putchar(c);
	}
	ungetc(c, fp);

	return r;
}

unsigned long get_long(FILE *fp)
{
	int i;
	unsigned long r;

	r = 0;
	for (i=0; i<4; i++) {
		r += ((unsigned long)putchar(getc(fp))) << (i * 8);
	}
	return r;
}

unsigned int get_int(FILE *fp)
{
	int i;
	unsigned int r;

	r = 0;
	for (i=0; i<2; i++) {
		r += ((unsigned int)putchar(getc(fp))) << (i * 8);
	}
	return r;
}

void error(char *s)
{
	fprintf(stderr, "%s\n", s);
	exit(1);
}

void usage()
{
	fprintf(stderr, "usage: %s [options] [file]\n", progname);
	fprintf(stderr, "  correct distortion of photo\n");
	fprintf(stderr, "  version 0.3\n");
	fprintf(stderr, "  option:\n");
	fprintf(stderr, "    -p param : set distortion parameter\n");
	fprintf(stderr, "               (-100 to 100 : default = 20)\n");
	fprintf(stderr, "    -f : convert full data\n");
	fprintf(stderr, "    -a : no anti-aliasing\n");
	fprintf(stderr, "  file: ppm / bmp\n");
	exit(1);
}

void exec(FILE *fp)
{
	int x, y;
	int i, j, c;
	double xcenter, ycenter;
	double srad, lrad2, co, xx, yy, co1, co2;
	unsigned char FAR_PTR *p;
	rgb color;
	unsigned long offset; /* for BMP header */
	int skip; /* for BMP padding */
#ifdef SAVE_MEMORY
	int y_max, y_min, y_min_tmp, y_tmp;
#endif


	c = putchar(getc(fp));
	if (c == 'P') { /* ppm */
		if (putchar(getc(fp)) != '6') error("file type error\n");
		skip_blank(fp);

		xsize = get_num(fp); skip_blank(fp);
		ysize = get_num(fp); skip_blank(fp);
		if (xsize == 0 || ysize == 0) error("file size error\n");

		while (1) {
			c = putchar(getc(fp));
			if (isspace(c)) break;
		}

		skip = 0;

	} else if (c == 'B') { /* BMP */

		if (putchar(getc(fp)) != 'M') error("file type error\n");
		for (i=0; i<8; i++) putchar(getc(fp));
		offset = get_long(fp);
		if (get_long(fp) == 12) { /* for OS/2 DIB */
			xsize = get_int(fp);
			ysize = get_int(fp);
			offset -= 26;
		} else {
			xsize = get_long(fp);
			ysize = get_long(fp);
			offset -= 30;
		}
		get_int(fp);
		if (get_int(fp) != 24) error("Not a 24-bit BMP\n");
		for (i=0; i<offset; i++) putchar(getc(fp));
		skip = 4 - (xsize * 3) % 4;
		if (skip == 4) skip = 0;
	} else {
		error("file type error\n");
	}


	xcenter=((double)xsize - 1.) / 2.;
	ycenter=((double)ysize - 1.) / 2.;

	srad = (xcenter > ycenter) ? ycenter : xcenter;
	lrad2 = xcenter * xcenter + ycenter * ycenter;
	co1 = param / 300. / lrad2;
	if (full == 1) {
		co2 = 1 - param / 300.;
	} else {
		co2 = 1 - param / 300. * (srad * srad / lrad2);
	}

	dat = (unsigned char FAR_PTR * FAR_PTR *)
	          malloc(sizeof(unsigned char FAR_PTR *) * ysize);
	if (dat == NULL) error("Not enough memory\n");
#ifdef SAVE_MEMORY
	y_max = -1; y_min = 0;
#else
	for (y=0; y<ysize; y++) {
		dat[y] = (unsigned char FAR_PTR *)
		             malloc(sizeof(unsigned char) * xsize * 3);
		if (dat[y] == NULL) error("Not enough memory\n");
		p = dat[y];
		for (x=0; x<xsize; x++) {
			for (i=0; i<3; i++) {
				if ((c = getc(fp)) == EOF) error("not enough data\n");
				*(p++) = c;
			}
		}
		for (i = 0; i < skip; i++) getc(fp); /* for BMP padding */
	}
#endif /* SAVE_MEMORY */

	for (y=0; y < ysize; y++) {
#ifdef SAVE_MEMORY
		y_min_tmp = ysize;
#endif
		for (x=0; x < xsize; x++) {
			co = (1 - co1*((x-xcenter)*(x-xcenter)+(y-ycenter)*(y-ycenter))) / co2;
			xx = (x - xcenter) * co + xcenter;
			yy = (y - ycenter) * co + ycenter;

#ifdef SAVE_MEMORY
			y_tmp = (int)yy;
			if (y_tmp < 0) y_tmp = 0;
			if (y_min_tmp > y_tmp) y_min_tmp = y_tmp;

			y_tmp = (int)yy + 1;
			if (y_tmp > ysize - 1) y_tmp = ysize - 1;
			if (y_tmp > y_max) {
				for (i = y_max + 1; i <= y_tmp; i++) {
					dat[i] = (unsigned char FAR_PTR *)
					             malloc(sizeof(unsigned char) * xsize * 3);
					if (dat[i] == NULL) error("Not enough memory\n");
#ifdef DEBUG
					fprintf(stderr, "malloc line %d\n", i);
#endif
					p = dat[i];
					for (j = 0; j < xsize * 3; j++) {
						if ((c = getc(fp)) == EOF) error("not enough data\n");
						*(p++) = c;
					}
					for (j = 0; j < skip; j++) getc(fp); /* for BMP padding */
				}
				y_max = y_tmp;
			}
#endif /* SAVE_MEMORY */

			color = get_rgb(xx, yy);
			putchar(color.r);
			putchar(color.g);
			putchar(color.b);
		} /* end of loop x */

		for (i = 0; i < skip; i++) putchar(0); /* for BMP padding */

#ifdef SAVE_MEMORY
		if (y_min_tmp > y_min) {
			for (i=y_min; i<y_min_tmp; i++) {
				free(dat[i]);
#ifdef DEBUG
				fprintf(stderr, "free line %d\n", i);
#endif
			}
			y_min = y_min_tmp;
		}
#endif /* SAVE_MEMORY */
	} /* end of loop y */

#ifdef SAVE_MEMORY
	for (y=y_min; y<ysize; y++) {
		free(dat[y]);
#ifdef DEBUG
		fprintf(stderr, "free line %d\n", y);
#endif
	}
#else /* SAVE_MEMORY */
	for (y=0; y<ysize; y++) {
		free(dat[y]);
	}
#endif /* SAVE_MEMORY */
	free(dat);
}


void main(int argc, char **argv)
{
	int i;
	FILE *fp;

#ifdef DOS
#ifdef LSI_C
	fsetbin(stdin);
	fsetbin(stdout);
#else
	setmode(fileno(stdin), O_BINARY);
	setmode(fileno(stdout), O_BINARY);
#endif
#endif /* DOS */

	progname = argv[0];

	fp = stdin;

	for (i=1; i<argc; i++) {
		if (argv[i][0] != '-') {
			if (i != argc-1) usage();
#ifdef DOS
			fp = fopen(argv[i], "rb");
#else
			fp = fopen(argv[i], "r");
#endif
			if (fp == NULL) {
				fprintf(stderr, "can\'t open %s\n", argv[i]);
				exit(1);
			}
			continue;
		}
		switch(argv[i][1]) {
		case 'p':
			i++;
			param = atof(argv[i]);
			break;
		case 'f':
			full = 1;
			break;
		case 'a':
			no_anti = 1;
			break;
		default:
			usage();
		}
	}

	if (param < 0) full = 1 - full;

	exec(fp);

	fclose(fp);
}
